/*
Eva library C++ (www.elxala.de)
(c) Copyright 2006 Alejandro Xalabarder Aulet

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef __EVALAYOUTMANAGER_H__
#define __EVALAYOUTMANAGER_H__

#include <FL/Fl_Widget.H>
#include "Eva.h"


/**
   (c) copyright 2006 Alejandro Xalabarder Aulet

   @author Alejandro Xalabarder
   @class  EvaLayoutManager
<pre>

   Table concept for layouting components with EvaLayoutManager:
   ------------------------------------------------------------

   With Evalayout manager all components are placed into a table where each column and row
   might be of one of following three types:

      - Fixed size. The width (columns) or height (rows) is given in pixels and remains always the same
      - Adaptable.  The width (columns) or height (rows) is taken from the component that begining
                    in this column or row needs the bigest size.
      - Expandable. On resizing the window or dialog these columns or rows receive the remaining
                    size of the window (total size - fixed size)

   The components might occupy more than one column and/or row.

   Format for EvaLayout's
   ---------------------------------------

   The number of columns and rows, the type of each one, the components used and its place into the table
   is called layout information and it is given to EvaLayout manager in a plain text format.

   The layout information has to be placed in a special structure, the Eva variable. More Eva variables
   might be loaded from a file in one step into an EvaUnit variable. This can be used for the layout of
   all dialogs in the application or for a "switchable" layout of a single dialog.

   The text format for a single Eva variable is

            EvaLayout, (horizontal margin), (vertical margin), (horizontal gap), (vertical gap)

            rows\columns,   (column header) , ..., (column header)
            (row header),   (component cell), ..., ...
            ...               ...           , ..., ...
            (row header),     ...           , ..., ...

   where
         "EvaLayout" is a constant text, margins and gaps are given in pixels
         "rows\columns" can be any text or empty string, it is irrelevant.

         (column header) and (row header)  :
            might be a character, an empty string or a number

            value  meaning
            -----  --------
              X    the column/row is expandable
              A    (or empty string) the column/row has to be resized acording to the
                   components default size (initial size)
           number  the pixels of the column/row

         (component cell) :

            represents the component that occupies the cell. It can be nothing (empty string),
            a component name or one of these two characters

            value  meaning
            -----  --------
              -    the component at the left of this symbol occupies also this cell
              +    the component above this symbol occupies also this cell


   For the file format each layout info has to be named using the characters < >

         <name of layout1>
               layout 1...
               ...
         <name of layout2>
               layout 2...
               ...

      ===================================
      EXAMPLE:

         let's say we want to lay-out the following form

         ----------------------------------------------
         |   label1  | text1     ------->   | button1 |
         ----------------------------------------------
         |   label2  | text2     ------->   | button2 |
         ------------|   |                  |----------
                     |   |                  | button3 |
                     |   V                  |----------
                     |                      |
                     ------------------------

         we can divide it in three columns and 4 rows (3 + 1 expandable) as follows

         <myFirstLayout>

            EvaLayout, 10, 10, 5, 5
            xxx,    A     ,    X    ,    A
              A, label1   , text1   ,  button1
              A, label2   , text2   ,  button2
              A,          ,   +     ,  button3
              X,          ,   +     ,

      ===================================

      This layoutInfo defines completly (except for the default size values of each component)
      the rules for positioning and resizing the components. We set this layoutInfo using
      the method 'setLayout'. It can be done more times with different layouts thus
      allowing dynamic layout switching.

      Of course EvaLayout need to know the physical components to be positioned and resized.
      This is done in the method 'addComponent'. This method has two parameters: the component
      name (that given in layoutInfo) and the windows component id (a number). Note that
      your programm has to create the physical components, the job of EvaLayoutManager is simply
      arrange them every time the window is resized or maximized.


   Coding example:
   -------------------------------------------------------------------

   Includes

      #include "EvaLayoutManager.h"
      #include "EvaLine.h"
      #include "EvaUnit.h"
      #include "EvaFile.h"

   Variables

      EvaLayoutManager vManager;

      Eva     oneLayInfo;
      EvaUnit moreLayInfo;


   Loading layout info


      oneLayInfo.addLine (EvaLine ("EvaLayout, 10, 10, 5, 5             "));
      oneLayInfo.addLine (EvaLine ("xxx,    A     ,    X    ,    A      "));
      oneLayInfo.addLine (EvaLine ("  A, label1   , text1   ,  button1  "));
      oneLayInfo.addLine (EvaLine ("  A, label2   , text2   ,  button2  "));
      oneLayInfo.addLine (EvaLine ("  A,          ,   +     ,  button3  "));
      oneLayInfo.addLine (EvaLine ("  X,          ,   +     ,           "));

      //
      // or loading it from a file
      //

      EvaFile eFile;
      eFile.load(moreLayInfo, "EvcLayouts.eva", "layouts");


   Setting layout info into the manager


      vManager.setLayout (layInfo);
      //
      // or
      //
      vManager.setLayout (moreLayInfo [0]);
      //
      // or
      //
      vManager.setLayout (moreLayInfo ["myLayout"]);


   Mapping physical components with the component names in layout info


      vManager.addComponent ("label1",  IDC_LABEL_1);
      vManager.addComponent ("label2",  IDC_LABEL_2);
      vManager.addComponent ("text1",   IDC_TEXT_1);

      //
      // or in MFC
      //

      vManager.addComponent("label1",  GetDlgItem (IDC_LABEL_1)->GetSafeHwnd());
      vManager.addComponent("label2",  GetDlgItem (IDC_LABEL_2)->GetSafeHwnd());
      vManager.addComponent("text1",   GetDlgItem (IDC_TEXT_1)->GetSafeHwnd());


   Recalculating the layout (on init or resize)

      // use in window messages WM_INITDIALOG and WM_SIZE and on switching the layout
      //
      void showLayout (HWND hDlg, bool forceInvalidate)
      {
         RECT rc;
         GetClientRect(hDlg, &rc);

         vManager.doLayout((int) (rc.right - rc.left), (int) (rc.bottom - rc.top));
      }

      //
      // or in MFC
      //

      // use in window messages ::OnInitDialog and ::OnSize and on switching the layout
      //
      void myDialog::showLayout ()
      {
         CRect rect;
         GetClientRect(&rect);

         gaston.doLayout(rect.Width(), rect.Height());
      }
*/

struct st_RECT {
    int left, right, top, bottom;
};

class EvaLayoutManager
{
public:

    EvaLayoutManager();

    /** sets the layout information given in an Eva variable
    */
    void setLayout (Eva layoutInfo);

    /**
        adds a component to be positioned and resized by EvaLayoutManager
        if the component is not referenced in the layout configuration then
        obviously it will not handled by EvaLayoutManager. A component with
        the same name should not be added twice without performing a call to
        removeComponents() in between.

        @param componentName logical name of the component used in layoutInfo. The
                name is case sensitive.
        @param componentId  windows id of the component. EvaLayoutManager will
                send positioning and resizing messages to the component using this id
    */
    void addComponent (const char * componentName, Fl_Widget *compWinHandle);

    /**
        removes all the previously added components. Note that the layoutInfo is not
        affected and therefore it is possible to add new or the same components again.
    */
    void removeComponents();

    /**
        makes the job of positioning and resizing all the components controlled by
        this layout manager. Typically to be called on WM_SIZE windows message.

       @params totWidth  total width of the dialog to be layed out
       @params totHeight total height of the dialog to be layed out
    */
    void doLayout(int aWidth, int aHeight);

    /**
        returns in the output variable 'vecNames' a string vector with the
        component names found in the layoutInfo. Note that this list of names
        comes from layoutInfo, that means,  we could have it before any addComponent
        is performed.
    */
    void getWidgets (std::vector<std::string> & vecNames);

    /** returns the number of columns of the layoutInfo
    */
    int nColumns ();

    /** returns the number of rows of the layoutInfo
    */
    int nRows ();

    /** returns the number of rows of the layoutInfo
    */
    std::string & widgetAt(int nrow, int ncol);


private:
    /// Structure to hold the needed information per component
    struct widgetInfo
    {
        widgetInfo ();
        widgetInfo (const char * wname, Fl_Widget *theHandle, st_RECT &originalRect);

        std::string name;      // name of the component in the layout array
        Fl_Widget   *Whand;     // component window handle

        bool   isLaidOut; // if the component has been found in the layout array
        // and therefore if indxPos is a valid calculated field
        st_RECT   indxPos;   // place in the layout array mesured in array indices
        // left  = column index
        // right = last column index
        // top   = row index
        // right = last row index

        st_RECT   iniRect;   // initial pos and size of the component
    };

    // variables for pre-calculated layout
    //
    bool isPrecalculated;

    int Hmargin;  // left and right margin
    int Vmargin;  // top and bottom margin
    int Hgap;     // distance between columns
    int Vgap;     // distance between rows
    int fijoH;    // intern precalculated not variable width
    int fijoV;    // intern precalculated not variable height


    std::vector<int> Hpos;      // calculated initial horizontal position
    std::vector<int> Hdim;      // calculated column width
    std::vector<int> HextraPos; // calculated on every paint

    std::vector<int> Vdim;      // calculated initial vertical position
    std::vector<int> Vpos;      // calculated row width
    std::vector<int> VextraPos; // calculated on every paint

    std::vector<int> columnsReparto;  // indexes of columns where the size need to be distributed
    std::vector<int> rowsReparto;     // indexes of rows where the size need to be distributed

    std::vector<widgetInfo> componentArray;   // components added through addComponent method

    /// layout information contained in a single Eva variable
    Eva evaLayout;

    /// precalculate the layout
    void precalculateAll ();

    // header of the column 'ncol'. Either "A", "X", "" or a number
    std::string & columnHeader (int ncol);

    // header of the row 'nrow'. Either "A", "X", "" or a number
    std::string & rowHeader    (int nrow);

    // minimum height of row 'nrow'
    int  minHeightOfRow   (int nrow);

    // minimum width of column 'ncol'
    int  minWidthOfColumn (int ncol);

    /**
       index at 'componentArray' of the component with name 'compName'
       returns -1 if the component is not found
    */
    int  indxComponent    (const std::string & compName);
};

// ----- inline functions
//

inline EvaLayoutManager::EvaLayoutManager()
    : isPrecalculated (false),
    Hmargin         (0),  // left and right margin
    Vmargin         (0),  // top and bottom margin
    Hgap            (0),  // distance between columns
    Vgap            (0),  // distance between rows
    fijoH           (0),  // intern precalculated not variable width
    fijoV           (0)   // intern precalculated not variable height
{
}

inline EvaLayoutManager::widgetInfo::widgetInfo ()
{
}

inline EvaLayoutManager::widgetInfo::widgetInfo (const char * wname, Fl_Widget* theHandle, st_RECT &originalRect)
    : name      (wname),
    Whand     (theHandle),
    isLaidOut (false),
    iniRect   (originalRect)
{
    indxPos.left = indxPos.right = indxPos.top = indxPos.bottom = 0;
}

inline int EvaLayoutManager::nColumns ()
{
    return std::max (0, evaLayout.cols(1) - 1);
}

inline int EvaLayoutManager::nRows ()
{
    return std::max (0, evaLayout.rows() - 2);
}

inline std::string & EvaLayoutManager::widgetAt (int nrow, int ncol)
{
    return evaLayout[nrow + 2][ncol + 1];
}

#endif
